"""Methods to load era5 data."""
__author__ = "Lukas Leufen"
__date__ = "2022-06-09"

import logging
import os
from functools import partial

import pandas as pd
import xarray as xr

from mlair import helpers
from mlair.configuration.era5_settings import era5_settings
from mlair.configuration.toar_data_v2_settings import toar_data_v2_settings
from mlair.helpers.data_sources.toar_data_v2 import load_station_information, combine_meta_data, correct_timezone
from mlair.helpers.data_sources.data_loader import EmptyQueryResult
from mlair.helpers.meteo import relative_humidity_from_dewpoint


def load_era5(station_name, stat_var, sampling, data_origin, time_dim, window_dim, target_dim, era5_data_path=None,
              era5_file_names=None):

    # make sure station_name parameter is a list
    station_name = helpers.to_list(station_name)

    # get data path
    data_path, file_names = era5_settings(sampling, era5_data_path=era5_data_path, era5_file_names=era5_file_names)

    # correct stat_var values if data is not aggregated (hourly)
    if sampling == "hourly":
        stat_var = {key: "values" for key in stat_var.keys()}
    else:
        raise ValueError(f"Given sampling {sampling} is not supported, only hourly sampling can be used.")

    # load station meta using toar-data v2 API
    meta_url_base, headers = toar_data_v2_settings("meta")
    station_meta = load_station_information(station_name, meta_url_base, headers)

    # sel data for station using sel method nearest
    logging.info(f"load data for {station_meta['codes'][0]} from ERA5")
    try:
        lon, lat = station_meta["coordinates"]["lng"], station_meta["coordinates"]["lat"]
        file_names = os.path.join(data_path, file_names)
        with xr.open_mfdataset(file_names, preprocess=partial(preprocess_era5_single_file, lon, lat)) as data:
            station_data = data.to_array().T.compute()
    except OSError as e:
        logging.info(f"Cannot load era5 data from path {data_path} and filenames {file_names} due to: {e}")
        return None, None

    if "relhum" in stat_var:
        relhum = relative_humidity_from_dewpoint(station_data.sel(variable="D2M"), station_data.sel(variable="T2M"))
        station_data = xr.concat([station_data, relhum.expand_dims({"variable": ["RHw"]})], dim="variable")
    station_data.coords["variable"] = _rename_era5_variables(station_data.coords["variable"].values)

    # check if all requested variables are available
    if set(stat_var).issubset(station_data.coords["variable"].values) is False:
        missing_variables = set(stat_var).difference(station_data.coords["variable"].values)
        origin = helpers.select_from_dict(data_origin, missing_variables)
        options = f"station={station_name}, origin={origin}"
        raise EmptyQueryResult(f"No data found for variables {missing_variables} and options {options} in JOIN.")
    else:
        station_data = station_data.sel(variable=list(stat_var.keys()))

    # convert to local timezone
    station_data.coords["time"] = correct_timezone(station_data.to_pandas(), station_meta, sampling).index
    station_data = station_data.rename({"time": time_dim, "variable": target_dim})

    # expand window_dim
    station_data = station_data.expand_dims({window_dim: [0]})

    # create meta data
    variable_meta = _emulate_meta_data(station_data.coords[target_dim].values)
    meta = combine_meta_data(station_meta, variable_meta)
    meta = pd.DataFrame.from_dict(meta, orient='index')
    meta.columns = station_name
    return station_data, meta


def preprocess_era5_single_file(lon, lat, ds):
    """Select lon and lat from data file and transform valid time into lead time."""
    ds = ds.sel(lon=lon, lat=lat, method="nearest", drop=True)
    return ds


def _emulate_meta_data(variables):
    general_meta = {"sampling_frequency": "hourly", "data_origin": "model", "data_origin_type": "model"}
    roles_meta = {"roles": [{"contact": {"organisation": {"name": "ERA5", "longname": "ECMWF"}}}]}
    variable_meta = {var: {"variable": {"name": var}, **roles_meta, ** general_meta} for var in variables}
    return variable_meta


def _rename_era5_variables(era5_names):
    mapper = {"SP": "press", "U10M": "u", "V10M": "v", "T2M": "temp", "D2M": "dew", "BLH": "pblheight",
              "TCC": "cloudcover", "RHw": "relhum"}
    era5_names = list(era5_names)
    try:
        join_names = list(map(lambda x: mapper[x], era5_names))
        return join_names
    except KeyError as e:
        raise KeyError(f"Cannot map names from era5 to join naming convention: {e}")