"""Functions to access join database."""
__author__ = 'Felix Kleinert, Lukas Leufen'
__date__ = '2019-10-16'

import datetime as dt
import logging
from typing import Iterator, Union, List, Dict, Tuple

import pandas as pd

from mlair import helpers
from mlair.configuration.join_settings import join_settings
from mlair.helpers.data_sources import toar_data, toar_data_v2, data_loader


# join_url_base = 'https://join.fz-juelich.de/services/rest/surfacedata/'
str_or_none = Union[str, None]


def download_join(station_name: Union[str, List[str]], stat_var: dict, station_type: str = None,
                  sampling: str = "daily", data_origin: Dict = None) -> [pd.DataFrame, pd.DataFrame]:
    """
    Read data from JOIN/TOAR.

    :param station_name: Station name e.g. DEBY122
    :param stat_var: key as variable like 'O3', values as statistics on keys like 'mean'
    :param station_type: set the station type like "traffic" or "background", can be none
    :param sampling: sampling rate of the downloaded data, either set to daily or hourly (default daily)
    :param data_origin: additional dictionary to specify data origin as key (for variable) value (origin) pair. Valid
        origins are "REA" for reanalysis data and "" (empty string) for observational data.

    :returns: data frame with all variables and statistics and meta data frame with all meta information
    """
    # make sure station_name parameter is a list
    station_name = helpers.to_list(station_name)

    # split network and origin information
    data_origin, network_name = split_network_and_origin(data_origin)

    # get data connection settings
    join_url_base, headers = join_settings(sampling)

    # load series information
    vars_dict, data_origin = load_series_information(station_name, station_type, network_name, join_url_base, headers,
                                                     data_origin, stat_var)

    # check if all requested variables are available
    if set(stat_var).issubset(vars_dict) is False:
        missing_variables = set(stat_var).difference(vars_dict)
        origin = helpers.select_from_dict(data_origin, missing_variables)
        options = f"station={station_name}, type={station_type}, network={network_name}, origin={origin}"
        raise data_loader.EmptyQueryResult(
            f"No data found for variables {missing_variables} and options {options} in JOIN.")

    # correct stat_var values if data is not aggregated (hourly)
    if sampling == "hourly":
        stat_var = {key: "values" for key in stat_var.keys()}

    # download all variables with given statistic
    data = None
    df = None
    meta = {}
    logging.info(f"load data for {station_name[0]} from JOIN")
    for var in _lower_list(sorted(vars_dict.keys())):
        if var in stat_var.keys():

            logging.debug('load: {}'.format(var))

            # create data link
            opts = {'base': join_url_base, 'service': 'stats', 'id': vars_dict[var], 'statistics': stat_var[var],
                    'sampling': sampling, 'capture': 0, 'format': 'json'}

            # load data
            data = data_loader.get_data(opts, headers)

            # adjust data format if given as list of list
            # no branch cover because this just happens when downloading hourly data using a secret token, not available
            # for CI testing.
            if isinstance(data, list):  # pragma: no branch
                data = correct_data_format(data)

            # correct namespace of statistics
            stat = data_loader.correct_stat_name(stat_var[var])

            # store data in pandas dataframe
            df = _save_to_pandas(df, data, stat, var)
            meta[var] = _correct_meta(data["metadata"])

            logging.debug('finished: {}'.format(var))

    if data:
        # load station meta using toar-data v2 API and convert to local timezone
        meta_url_base, headers = toar_data_v2.toar_data_v2_settings("meta")
        station_meta = toar_data_v2.load_station_information(station_name, meta_url_base, headers)
        df = toar_data_v2.correct_timezone(df, station_meta, sampling)

        # create meta data
        meta = toar_data_v2.combine_meta_data(station_meta, meta)
        meta = pd.DataFrame.from_dict(meta, orient='index')
        meta.columns = station_name
        return df, meta
    else:
        raise data_loader.EmptyQueryResult("No data found in JOIN.")


def _correct_meta(meta):
    meta_out = {}
    for k, v in meta.items():
        if k.startswith("station"):
            _k = k.split("_", 1)[1]
            _d = meta_out.get("station", {})
            _d[_k] = v
            meta_out["station"] = _d
        elif k.startswith("parameter"):
            _k = k.split("_", 1)[1]
            _d = meta_out.get("variable", {})
            _d[_k] = v
            meta_out["variable"] = _d
        elif k == "network_name":
            if v == "AIRBASE":
                _d = {"name": "EEA", "longname": "European Environment Agency", "kind": "government"}
            elif v == "UBA":
                _d = {"name": "UBA", "longname": "Umweltbundesamt", "kind": "government", "country": "Germany"}
            else:
                _d = {"name": v}
            meta_out["roles"] = [{"contact": {"organisation": _d}}]
        elif k in ["google_resolution", "numid"]:
            continue
        else:
            meta_out[k] = v
    return meta_out


def split_network_and_origin(origin_network_dict: dict) -> Tuple[Union[None, dict], Union[None, dict]]:
    """
    Split given dict into network and data origin.

    Method is required to transform Toar-Data v2 structure (using only origin) into Toar-Data v1 (JOIN) structure (which
    uses origin and network parameter). Furthermore, EEA network (v2) is renamed to AIRBASE (v1).
    """
    if origin_network_dict is None or len(origin_network_dict) == 0:
        data_origin, network = None, None
    else:
        data_origin = {}
        network = {}
        for k, v in origin_network_dict.items():
            network[k] = []
            for _network in helpers.to_list(v):
                if _network.lower() == "EEA".lower():
                    network[k].append("AIRBASE")
                elif _network.lower() != "REA".lower():
                    network[k].append(_network)
            if "REA" in v:
                data_origin[k] = "REA"
            else:
                data_origin[k] = ""
            network[k] = filter_network(network[k])
    return data_origin, network


def filter_network(network: list) -> Union[list, None]:
    """
    Filter given list of networks.

    :param network: list of various network names (can contain duplicates)
    :return: sorted list with unique entries
    """
    sorted_network = []
    for v in list(filter(lambda x: x != "", network)):
        if v not in sorted_network:
            sorted_network.append(v)
    if len(sorted_network) == 0:
        sorted_network = None
    return sorted_network


def correct_data_format(data):
    """
    Transform to the standard data format.

    For some cases (e.g. hourly data), the data is returned as list instead of a dictionary with keys datetime, values
    and metadata. This functions addresses this issue and transforms the data into the dictionary version.

    :param data: data in hourly format

    :return: the same data but formatted to fit with aggregated format
    """
    formatted = {"datetime": [],
                 "values": [],
                 "metadata": data[-1]}
    for d in data[:-1]:
        for k, v in zip(["datetime", "values"], d):
            formatted[k].append(v)
    return formatted


def load_series_information(station_name: List[str], station_type: str_or_none, network_name: str_or_none,
                            join_url_base: str, headers: Dict, data_origin: Dict = None, stat_var: Dict = None) -> [Dict, Dict]:
    """
    List all series ids that are available for given station id and network name.

    :param station_name: Station name e.g. DEBW107
    :param station_type: station type like "traffic" or "background"
    :param network_name: measurement network of the station like "UBA" or "AIRBASE"
    :param join_url_base: base url name to download data from
    :param headers: additional headers information like authorization, can be empty
    :param data_origin: additional information to select a distinct series e.g. from reanalysis (REA) or from observation
        ("", empty string). This dictionary should contain a key for each variable and the information as key
    :return: all available series for requested station stored in an dictionary with parameter name (variable) as key
        and the series id as value.
    """
    network_name_opts = _create_network_name_opts(network_name)
    parameter_name_opts = _create_parameter_name_opts(stat_var)
    opts = {"base": join_url_base, "service": "search", "station_id": station_name[0], "station_type": station_type,
            "network_name": network_name_opts, "as_dict": "true", "parameter_name": parameter_name_opts,
            "columns": "id,network_name,station_id,parameter_name,parameter_label,parameter_attribute"}
    station_vars = data_loader.get_data(opts, headers)
    logging.debug(f"{station_name}: {station_vars}")
    return _select_distinct_series(station_vars, data_origin, network_name)


def _create_parameter_name_opts(stat_var):
    if stat_var is None:
        parameter_name_opts = None
    else:
        parameter_name_opts = ",".join(stat_var.keys())
    return parameter_name_opts


def _create_network_name_opts(network_name):
    if network_name is None:
        network_name_opts = network_name
    elif isinstance(network_name, list):
        network_name_opts = ",".join(helpers.to_list(network_name))
    elif isinstance(network_name, dict):
        _network = []
        for v in network_name.values():
            _network.extend(helpers.to_list(v))
        network_name_opts = ",".join(filter(lambda x: x is not None, set(_network)))
        network_name_opts = None if len(network_name_opts) == 0 else network_name_opts
    else:
        raise TypeError(f"network_name parameter must be of type None, list, or dict. Given is {type(network_name)}.")
    return network_name_opts


def _select_distinct_series(vars: List[Dict], data_origin: Dict = None, network_name: Union[str, List[str]] = None) \
        -> [Dict, Dict]:
    """
    Select distinct series ids for all variables. Also check if a parameter is from REA or not.
    """
    data_origin = {} if data_origin is None else data_origin
    selected, data_origin = _select_distinct_data_origin(vars, data_origin)

    network_name = [] if network_name is None else network_name
    selected = _select_distinct_network(selected, network_name)

    # extract id
    selected = {k: v["id"] for k, v in selected.items()}
    return selected, data_origin


def _select_distinct_network(vars: dict, network_name: Union[list, dict]) -> dict:
    """
    Select distinct series regarding network name. The order the network names are provided in parameter `network_name`
    indicates priority (from high to low). If no network name is provided, first entry is used and a logging info is
    issued. In case network names are given but no match can be found, this method raises a ValueError.

    :param vars: dictionary with all series candidates already grouped by variable name as key. Value should be a list
        of possible candidates to select from. Each candidate must be a dictionary with at least keys `id` and
        `network_name`.
    :param network_name: list of networks to use with increasing priority (1st element has priority). Can be empty list
        indicating to use always first candidate for each variable.
    :return: dictionary with single series reference for each variable
    """
    if isinstance(network_name, (list, str)):
        network_name = {var: helpers.to_list(network_name) for var in vars.keys()}
    selected = {}
    for var, series in vars.items():
        res = []
        network_list = helpers.to_list(network_name.get(var, []) or [])
        for network in network_list:
            res.extend(list(filter(lambda x: x["network_name"].upper() == network.upper(), series)))
        if len(res) > 0:  # use first match which has the highest priority
            selected[var] = res[0]
        else:
            if len(network_list) == 0:  # just print message which network is used if none is provided
                selected[var] = series[0]
                logging.info(f"Could not find a valid match for variable {var} and networks {network_name.get(var, [])}"
                             f"! Therefore, use first answer from JOIN: {series[0]}")
            else:  # raise error if network name is provided but no match could be found
                raise ValueError(f"Cannot find a valid match for requested networks {network_name.get(var, [])} and "
                                 f"variable {var} as only following networks are available in JOIN: "
                                 f"{list(map(lambda x: x['network_name'], series))}")
    return selected


def _select_distinct_data_origin(vars: List[Dict], data_origin: Dict) -> (Dict[str, List], Dict):
    """
    Select distinct series regarding their data origin. Series are grouped as list according to their variable's name.
    As series can be reported with different network attribution, results might contain multiple entries for a variable.
    This method assumes the default data origin for chemical variables as `` (empty source) and for meteorological
    variables as `REA`.
    :param vars: list of all entries to check data origin for
    :param data_origin: data origin to match series with, if empty default values are used
    :return: dictionary with unique variable names as keys and list of respective series as values
    """
    data_origin_default = {"cloudcover": "REA", "humidity": "REA", "pblheight": "REA", "press": "REA", "relhum": "REA",
                           "temp": "REA", "totprecip": "REA", "u": "REA", "v": "REA",
                           "no": "", "no2": "", "o3": "", "pm10": "", "so2": ""}
    selected = {}
    for var in vars:
        name = var["parameter_name"].lower()
        var_attr = var["parameter_attribute"].lower()
        if name not in data_origin.keys():
            data_origin.update({name: data_origin_default.get(name, "")})
        attr = data_origin.get(name, "").lower()
        if var_attr == attr:
            selected[name] = selected.get(name, []) + helpers.to_list(var)
    return selected, data_origin


def _save_to_pandas(df: Union[pd.DataFrame, None], data: dict, stat: str, var: str) -> pd.DataFrame:
    """
    Save given data in data frame.

    If given data frame is not empty, the data is appened as new column.

    :param df: data frame to append the new data, can be none
    :param data: new data to append or format as data frame containing the keys 'datetime' and '<stat>'
    :param stat: extracted statistic to get values from data (e.g. 'mean', 'dma8eu')
    :param var: variable the data is from (e.g. 'o3')

    :return: new created or concatenated data frame
    """
    if len(data["datetime"][0]) == 19:
        str_format = "%Y-%m-%d %H:%M:%S"
    else:
        str_format = "%Y-%m-%d %H:%M"
    index = map(lambda s: dt.datetime.strptime(s, str_format), data['datetime'])
    if df is None:
        df = pd.DataFrame(data[stat], index=index, columns=[var])
    else:
        df = pd.concat([df, pd.DataFrame(data[stat], index=index, columns=[var])], axis=1)
    return df


def _lower_list(args: List[str]) -> Iterator[str]:
    """
    Lower all elements of given list.

    :param args: list with string entries to lower

    :return: iterator that lowers all list entries
    """
    for string in args:
        yield string.lower()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    var_all_dic = {'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                   'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                   'pblheight': 'maximum'}
    station = 'DEBW107'
    # download_join(station, var_all_dic, sampling="daily")
    download_join(station, var_all_dic, sampling="hourly")
